from datetime import datetime
from pathlib import Path

from pytz import UTC

from metric_config_parser.analysis import AnalysisSpec
from metric_config_parser.config import ConfigCollection
from metric_config_parser.experiment import Channel, Experiment
from metric_config_parser.metric import AnalysisPeriod


class TestConfigIntegration:
    def test_overall_retention_regression(self):
        config_collection = ConfigCollection.from_github_repos(
            [
                "https://github.com/mozilla/metric-hub",
                "https://github.com/mozilla/metric-hub/tree/main/jetstream",
            ]
        )
        experiment_slug = "ios-onboarding-search-widget"
        config_collection.as_of(datetime.fromisoformat("2023-11-16T21:44:49+00:00"))
        experiment = Experiment(
            experimenter_slug=None,
            normandy_slug=experiment_slug,
            type="v6",
            status="Complete",
            branches=["control", "treatment-a", "treatment-b"],
            reference_branch="control",
            is_high_population=False,
            start_date=datetime(2023, 9, 8),
            proposed_enrollment=14,
            enrollment_end_date=datetime(2023, 9, 19),
            end_date=datetime(2023, 10, 16),
            app_name="firefox_ios",
            channel=Channel.RELEASE,
            is_enrollment_paused=True,
            outcomes=["onboarding", "default_browser"],
        )
        spec = AnalysisSpec.default_for_experiment(experiment, config_collection)
        experiment_config = spec.resolve(experiment, config_collection)

        overall_metric_names = [
            summary.metric.name for summary in experiment_config.metrics[AnalysisPeriod.OVERALL]
        ]
        assert "retained" not in overall_metric_names
        assert "opened_as_default" in overall_metric_names
        assert "default_browser_card_go_to_settings_pressed" in overall_metric_names

        weekly_metric_names = [
            summary.metric.name for summary in experiment_config.metrics[AnalysisPeriod.WEEK]
        ]
        assert "retained" in weekly_metric_names
        assert "opened_as_default" in weekly_metric_names
        assert "default_browser_card_go_to_settings_pressed" in weekly_metric_names

    def test_configs_from_repo(self):
        config_collection = ConfigCollection.from_github_repos(
            ["https://github.com/mozilla/metric-hub"]
        )
        assert config_collection is not None
        assert config_collection.get_platform_defaults("firefox_desktop") is None
        assert config_collection.spec_for_outcome("test", "firefox_desktop") is None
        assert (
            config_collection.get_data_source_definition("clients_daily", "firefox_desktop")
            is not None
        )
        assert config_collection.get_metric_definition("baseline_ping_count", "fenix") is not None
        assert config_collection.get_metric_definition("not_exist", "firefox_desktop") is None
        assert (
            config_collection.get_segment_definition("regular_users_v3", "firefox_desktop")
            is not None
        )
        assert config_collection.get_segments_for_app("firefox_desktop") is not None

    def test_configs_from_multiple_repos(self):
        config_collection = ConfigCollection.from_github_repos(
            repo_urls=[ConfigCollection.repo_url, ConfigCollection.repo_url]
        )
        assert config_collection is not None
        assert config_collection.functions is not None

        default_collection = ConfigCollection.from_github_repo()
        assert len(config_collection.configs) == len(default_collection.configs)
        assert config_collection.outcomes == default_collection.outcomes
        assert len(config_collection.defaults) == len(default_collection.defaults)
        assert len(config_collection.definitions) == len(default_collection.definitions)

    def test_config_from_repo_tree(self):
        config_collection = ConfigCollection.from_github_repo(
            "https://github.com/mozilla/metric-hub/tree/main/jetstream"
        )
        assert config_collection.configs is not None

    def test_config_from_repo_tree_multiple(self):
        config_collection = ConfigCollection.from_github_repos(
            repo_urls=[
                ConfigCollection.repo_url,
                "https://github.com/mozilla/metric-hub/tree/main/jetstream",
            ]
        )

        assert config_collection.configs is not None
        assert len(config_collection.definitions) > 0

    def test_config_collection_from_branch(self):
        config_collection = ConfigCollection.from_github_repos(
            [
                "https://github.com/mozilla/metric-hub/tree/main",
                "https://github.com/mozilla/metric-hub/tree/main/jetstream/",
            ],
        )
        assert len(config_collection.configs) > 0

    def test_config_as_of(self):
        config_collection = ConfigCollection.from_github_repo(
            "https://github.com/mozilla/metric-hub/tree/main/jetstream"
        ).as_of(
            UTC.localize(datetime(2023, 5, 15))
        )  # 6a052aea23e7e2332a20c992b2e6f07468c3d161

        assert config_collection is not None
        assert config_collection.spec_for_outcome("networking", "firefox_desktop") is None

        config_collection = config_collection.as_of(UTC.localize(datetime(2023, 5, 30)))  # 0f92ef5
        assert config_collection.spec_for_outcome("networking", "firefox_desktop") is not None

    def test_config_as_of_multiple_repos(self):
        config_collection = ConfigCollection.from_github_repos(
            repo_urls=[
                ConfigCollection.repo_url,
                "https://github.com/mozilla/metric-hub/tree/main/jetstream",
            ]
        )

        assert config_collection is not None
        assert config_collection.spec_for_outcome("networking", "firefox_desktop") is not None
        assert config_collection.get_metric_definition("daily_active_users_v2", "fenix") is not None

        config_collection = config_collection.as_of(
            UTC.localize(datetime(2023, 5, 15))
        )  # 6a052aea23e7e2332a20c992b2e6f07468c3d161

        assert config_collection.spec_for_outcome("networking", "firefox_desktop") is None
        assert config_collection.get_metric_definition("daily_active_users_v2", "fenix") is None

        config_collection = config_collection.as_of(
            UTC.localize(datetime(2023, 5, 25, 20, 0, 0))  # 2fa5433
        )

        assert config_collection.spec_for_outcome("networking", "firefox_desktop") is None
        assert config_collection.get_metric_definition("daily_active_users_v2", "fenix") is not None

    def test_remove_tmp_dir_on_destruct(self):
        config_collection = ConfigCollection.from_github_repos(
            repo_urls=[
                ConfigCollection.repo_url,
                "https://github.com/mozilla/metric-hub/tree/main/jetstream",
            ]
        )

        tmp_dirs = [Path(r.repo.git_dir).parent for r in config_collection.repos]

        for tmp_dir in tmp_dirs:
            assert tmp_dir.exists()

        del config_collection

        for tmp_dir in tmp_dirs:
            assert tmp_dir.exists() is False



from typing import TYPE_CHECKING, Dict, Optional

import attr

if TYPE_CHECKING:
    from metric_config_parser.config import ConfigCollection
    from metric_config_parser.monitoring import MonitoringSpec
    from metric_config_parser.project import ProjectConfiguration

from metric_config_parser.data_source import DataSource, DataSourceReference
from metric_config_parser.util import converter


@attr.s(auto_attribs=True)
class Dimension:
    """Represents a dimension for segmenting client populations."""

    name: str
    data_source: DataSource
    select_expression: str
    friendly_name: Optional[str] = None
    description: Optional[str] = None


@attr.s(auto_attribs=True)
class DimensionDefinition:
    """Describes the interface for defining a dimension in configuration."""

    name: str  # implicit in configuration
    select_expression: str
    data_source: DataSourceReference
    friendly_name: Optional[str] = None
    description: Optional[str] = None

    def resolve(
        self,
        spec: "MonitoringSpec",
        conf: "ProjectConfiguration",
        configs: "ConfigCollection",
    ) -> Dimension:
        """Create and return a `Dimension` from the definition."""
        return Dimension(
            name=self.name,
            data_source=self.data_source.resolve(spec, conf, configs),
            select_expression=self.select_expression,
            friendly_name=self.friendly_name,
            description=self.description,
        )

    def merge(self, other: "DimensionDefinition"):
        """Merge with another dimension definition."""
        for key in attr.fields_dict(type(self)):
            setattr(self, key, getattr(other, key) or getattr(self, key))


@attr.s(auto_attribs=True)
class DimensionsSpec:
    """Describes the interface for defining custom dimensions."""

    definitions: Dict[str, DimensionDefinition] = attr.Factory(dict)

    @classmethod
    def from_dict(cls, d: dict) -> "DimensionsSpec":
        """Create a `DimensionsSpec` from a dictionary."""
        d = dict((k.lower(), v) for k, v in d.items())

        definitions = {
            k: converter.structure({"name": k, **v}, DimensionDefinition) for k, v in d.items()
        }
        return cls(definitions=definitions)

    def merge(self, other: "DimensionsSpec"):
        """
        Merge another dimension spec into the current one.

        The `other` DimensionsSpec overwrites existing keys.
        """
        seen = []
        for key, _ in self.definitions.items():
            if key in other.definitions:
                self.definitions[key].merge(other.definitions[key])
            seen.append(key)
        for key, definition in other.definitions.items():
            if key not in seen:
                self.definitions[key] = definition


converter.register_structure_hook(DimensionsSpec, lambda obj, _type: DimensionsSpec.from_dict(obj))


@attr.s(auto_attribs=True)
class DimensionReference:
    """Represents a reference to a dimension."""

    name: str

    def resolve(
        self,
        spec: "MonitoringSpec",
        conf: "ProjectConfiguration",
        configs: "ConfigCollection",
    ) -> Dimension:
        """Return the referenced `Dimension`."""
        if self.name in spec.dimensions.definitions:
            return spec.dimensions.definitions[self.name].resolve(spec, conf, configs)
        raise ValueError(f"Could not locate dimension {self.name}")


converter.register_structure_hook(
    DimensionReference, lambda obj, _type: DimensionReference(name=obj)
)



from datetime import datetime
from textwrap import dedent

import pytz
import toml

from metric_config_parser.monitoring import MonitoringSpec
from metric_config_parser.project import MonitoringPeriod


class TestPopulationSpec:
    def test_overwrite_population(self, config_collection):
        config_str = dedent(
            """
            [project]
            name = "foo"
            xaxis = "build_id"
            metrics = []
            start_date = "2022-01-01"
            end_date = "2022-02-01"

            [project.population]
            data_source = "foo"
            boolean_pref = "TRUE"
            branches = ["treatment"]
            dimensions = ["os"]
            group_by_dimension = "os"

            [data_sources]
            [data_sources.foo]
            from_expression = "test"

            [dimensions]
            [dimensions.os]
            select_expression = "os"
            data_source = "foo"
            """
        )

        spec = MonitoringSpec.from_dict(toml.loads(config_str))

        config_str = dedent(
            """
            [project]
            name = "foo bar"
            end_date = "2022-03-01"
            skip_default_metrics = true

            [project.population]
            boolean_pref = "FALSE"
            branches = ["test-1"]
            """
        )

        spec2 = MonitoringSpec.from_dict(toml.loads(config_str))
        spec.merge(spec2)
        cfg = spec.resolve(experiment=None, configs=config_collection)

        assert cfg.project.name == "foo bar"
        assert cfg.project.xaxis == MonitoringPeriod.BUILD_ID
        assert cfg.project.start_date == datetime(2022, 1, 1, tzinfo=pytz.utc)
        assert cfg.project.end_date == datetime(2022, 3, 1, tzinfo=pytz.utc)
        assert cfg.project.population.data_source.name == "foo"
        assert cfg.project.population.boolean_pref == "FALSE"
        assert cfg.project.population.branches == ["treatment"]
        assert cfg.project.skip_default_metrics
        assert len(cfg.dimensions) == 1



from collections import defaultdict
from typing import Any, Dict, Mapping, Optional, Union

import attr

from .errors import InvalidConfigurationException
from .util import converter


@attr.s(auto_attribs=True)
class ParameterDefinition:
    """
    * distinct_by_branch * indicates whether value for the parameter
        needs to be specified for each branch
        or if the value can be applied across all configuration branches (default: False)
    """

    name: str  # implicit in configuration
    friendly_name: Optional[str] = None
    description: Optional[str] = None
    value: Optional[Union[str, Dict[str, str]]] = None
    distinct_by_branch: Optional[bool] = False
    default: Optional[Union[str, Dict[str, Any]]] = None

    def validate(self) -> "ParameterDefinition":
        """
        Validates that branch related configuration is correct.

        It should not run on every instance of ParameterDefinition object.
        Outcome parameters containing defaults would not always adhere to
        the rules defined here.
        """

        if self.distinct_by_branch and not (
            isinstance(self.value, dict) or isinstance(self.default, dict)
        ):
            error_msg = (
                f"Parameter {self.name} configured "
                "to be distinct by branch, a mapping expected in the following format: "
                'for values: value.branch_1 = "1"'
                'and defaults: default.branch_1 = "1"'
                "See https://experimenter.info/jetstream/outcomes#parameterizing-outcomes for more information"  # noqa: E501
            )
            raise InvalidConfigurationException(error_msg)

        elif not self.distinct_by_branch and not (
            isinstance(self.value, str) or isinstance(self.default, str)
        ):
            error_msg = (
                f"Parameter {self.name} configured "
                "to not be distinct by branch, but wrong value type provided. "
                "Expected format: "
                f'value = "param_value", provided: {self.value}'
                f'default = "", provided: {self.default}'
                "See https://experimenter.info/jetstream/outcomes#parameterizing-outcomes for more information"  # noqa: E501
            )
            raise InvalidConfigurationException(error_msg)

        return self


converter.register_structure_hook(
    ParameterDefinition, lambda obj, _type: ParameterDefinition(**obj)
)


@attr.s(auto_attribs=True)
class ParameterSpec:
    """
    Object for holding definitions of all parameters.
    """

    definitions: Dict[str, ParameterDefinition] = attr.Factory(dict)

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> "ParameterSpec":
        """
        Converts a dictionary object containing parameter configuration
        into a ParameterSpec Object that contains a ParameterDefinition
        for each parameter.

        """

        params: Dict[str, Any] = {"definitions": defaultdict()}

        for param_name, param_config in d.items():
            params["definitions"][param_name] = converter.structure(
                {
                    "name": param_name,
                    **dict((kk.lower(), vv) for kk, vv in param_config.items()),
                },
                ParameterDefinition,
            )

        return cls(**params)


converter.register_structure_hook(ParameterSpec, lambda obj, _type: ParameterSpec.from_dict(obj))



import copy
import fnmatch
import re
from collections import defaultdict
from enum import Enum
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import attr
import jinja2
from mozilla_nimbus_schemas.jetstream import AnalysisBasis

from metric_config_parser.errors import DefinitionNotFound

if TYPE_CHECKING:
    from .analysis import AnalysisSpec
    from .config import ConfigCollection
    from .experiment import ExperimentConfiguration
    from .definition import DefinitionSpecSub
    from .project import ProjectConfiguration

from . import AnalysisUnit
from .data_source import DataSource, DataSourceReference
from .parameter import ParameterDefinition
from .pre_treatment import PreTreatmentReference
from .statistic import Statistic
from .util import converter, is_valid_slug


class AnalysisPeriod(Enum):
    DAY = "day"
    WEEK = "week"
    DAYS_28 = "days28"
    OVERALL = "overall"
    PREENROLLMENT_WEEK = "preenrollment_week"
    PREENROLLMENT_DAYS_28 = "preenrollment_days28"

    @property
    def mozanalysis_label(self) -> str:
        d = {
            "day": "daily",
            "week": "weekly",
            "days28": "28_day",
            "overall": "overall",
            "preenrollment_week": "preenrollment_weekly",
            "preenrollment_days28": "preenrollment_days28",
        }
        return d[self.value]

    @property
    def table_suffix(self) -> str:
        d = {
            "day": "daily",
            "week": "weekly",
            "days28": "days28",
            "overall": "overall",
            "preenrollment_week": "preenrollment_weekly",
            "preenrollment_days28": "preenrollment_days28",
        }
        return d[self.value]


class MetricLevel(Enum):
    GOLD = "gold"
    SILVER = "silver"
    BRONZE = "bronze"


@attr.s(auto_attribs=True)
class Summary:
    """Represents a metric with a statistical treatment."""

    metric: "Metric"
    statistic: "Statistic"
    pre_treatments: List[PreTreatmentReference] = attr.Factory(list)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Metric:
    """
    Metric representation.

    Metrics are supersets of mozanalysis metrics with additional
    metadata required for analysis.
    """

    name: str
    data_source: Optional[DataSource]
    select_expression: Optional[str]
    friendly_name: Optional[str] = None
    description: Optional[str] = None
    bigger_is_better: bool = True
    analysis_bases: List[AnalysisBasis] = [
        AnalysisBasis.ENROLLMENTS,
        AnalysisBasis.EXPOSURES,
    ]
    type: str = "scalar"
    category: Optional[str] = None
    depends_on: Optional[List[Summary]] = None
    owner: Optional[List[str]] = None
    deprecated: bool = False
    level: Optional[MetricLevel] = None
    analysis_units: List[AnalysisUnit] = [AnalysisUnit.CLIENT]


@attr.s(auto_attribs=True)
class MetricReference:
    name: str

    def resolve(
        self,
        spec: "AnalysisSpec",
        conf: Union["ExperimentConfiguration", "ProjectConfiguration"],
        configs: "ConfigCollection",
    ) -> List[Summary]:
        if self.name in spec.metrics.definitions:
            return spec.metrics.definitions[self.name].resolve(spec, conf, configs)

        metric_definition = configs.get_metric_definition(self.name, conf.app_name)
        if metric_definition:
            return metric_definition.resolve(spec, conf, configs=configs)

        raise DefinitionNotFound(f"Could not locate metric {self.name}")


# These are bare strings in the configuration file.
converter.register_structure_hook(MetricReference, lambda obj, _type: MetricReference(name=obj))

converter.register_structure_hook(Union[str, List[str], None], lambda obj, _type: obj)


@attr.s(auto_attribs=True)
class MetricDefinition:
    """Describes the interface for defining a metric in configuration.

    The `select_expression` of the metric may use Jinja2 template syntax to refer to the
    aggregation helper functions defined in `mozanalysis.metrics`, like
        '{{agg_any("payload.processes.scalars.some_boolean_thing")}}'
    """

    name: str  # implicit in configuration
    statistics: Optional[Dict[str, Dict[str, Any]]] = None
    select_expression: Optional[str] = None
    data_source: Optional[DataSourceReference] = None
    friendly_name: Optional[str] = None
    description: Optional[str] = None
    bigger_is_better: bool = True
    analysis_bases: Optional[List[AnalysisBasis]] = None
    type: Optional[str] = None
    category: Optional[str] = None
    depends_on: Optional[List[MetricReference]] = None
    owner: Optional[Union[str, List[str]]] = None
    deprecated: bool = False
    level: Optional[MetricLevel] = None
    analysis_units: Optional[List[AnalysisUnit]] = None

    @staticmethod
    def generate_select_expression(
        param_definitions: Dict[str, ParameterDefinition],
        select_expr_template: Union[str, jinja2.nodes.Template],
        configs: "ConfigCollection",
    ) -> str:
        """
        Takes in param configuration and converts it to a select statement string
        """

        if "parameters" not in str(select_expr_template):
            return configs.get_env().from_string(select_expr_template).render()

        formatted_params: Dict[str, Any] = defaultdict()

        for param_name, param_definition in param_definitions.items():
            if param_definition.distinct_by_branch and isinstance(param_definition.value, dict):
                formatted_params.update(
                    {
                        param_name: "CASE e.branch "
                        + " ".join(
                            [
                                f'WHEN "{branch}" THEN "{value}"'
                                for branch, value in param_definition.value.items()
                            ]
                        )
                        + " END"
                    }
                )
            else:
                formatted_params.update({param_name: param_definition.value})

        return (
            configs.get_env().from_string(select_expr_template).render(parameters=formatted_params)
        )

    def resolve(
        self,
        spec: "DefinitionSpecSub",
        conf: Union["ExperimentConfiguration", "ProjectConfiguration"],
        configs: "ConfigCollection",
    ) -> List[Summary]:
        metric_summary = None
        metric = None
        upstream_metrics = None

        # check if metric depends on other metrics
        if self.depends_on:
            upstream_metrics = []
            # resolve upstream metrics
            for metric_ref in self.depends_on:
                # check if upstream metric is defined externally as a "definition"
                upstream_metric = configs.get_metric_definition(metric_ref.name, conf.app_name)

                if upstream_metric is None:
                    # check if upstream metric is part of the analysis spec
                    upstream_metric = spec.metrics.definitions.get(metric_ref.name, None)

                if upstream_metric is None:
                    raise DefinitionNotFound(
                        f"No definition found for referenced upstream metric {metric_ref}"
                    )

                upstream_metrics += upstream_metric.resolve(spec, conf, configs)

        if self.select_expression is None or self.data_source is None:
            # checks if a metric from mozanalysis was referenced
            metric_definition = configs.get_metric_definition(self.name, conf.app_name)

            if metric_definition is None and upstream_metrics is None:
                raise DefinitionNotFound(
                    f"No default definition found for referenced metric {self.name}"
                )
            elif upstream_metrics:
                metric = Metric(
                    name=self.name,
                    data_source=None,
                    select_expression=None,
                    friendly_name=(
                        dedent(self.friendly_name) if self.friendly_name else self.friendly_name
                    ),
                    description=(
                        dedent(self.description) if self.description else self.description
                    ),
                    bigger_is_better=self.bigger_is_better,
                    analysis_bases=self.analysis_bases
                    or [AnalysisBasis.ENROLLMENTS, AnalysisBasis.EXPOSURES],
                    type=self.type or "scalar",
                    category=self.category,
                    depends_on=upstream_metrics,
                    owner=[self.owner] if isinstance(self.owner, str) else self.owner,
                    deprecated=self.deprecated,
                    level=self.level,
                    analysis_units=self.analysis_units or [AnalysisUnit.CLIENT],
                )
            elif metric_definition:
                metric_definition.analysis_bases = (
                    self.analysis_bases
                    or metric_definition.analysis_bases
                    or [
                        AnalysisBasis.ENROLLMENTS,
                        AnalysisBasis.EXPOSURES,
                    ]
                )
                metric_definition.statistics = self.statistics
                metric_definition.analysis_units = (
                    self.analysis_units or metric_definition.analysis_units or [AnalysisUnit.CLIENT]
                )
                metric_summary = metric_definition.resolve(spec, conf, configs)
        else:
            select_expression = self.generate_select_expression(
                spec.parameters.definitions,
                select_expr_template=self.select_expression,
                configs=configs,
            )

            # ensure all of metric's analysis_units are supported by data_source
            resolved_ds = self.data_source.resolve(spec, conf, configs)
            analysis_units = self.analysis_units or [AnalysisUnit.CLIENT]
            for agg_unit in analysis_units:
                if agg_unit not in resolved_ds.analysis_units:
                    raise ValueError(
                        f"data_source {resolved_ds.name} does not support "
                        f"all analysis_units specified by metric {self.name}: "
                        f"analysis_units for metric: {analysis_units}, "
                        f"analysis_units for data_source: {resolved_ds.analysis_units}"
                    )

            metric = Metric(
                name=self.name,
                data_source=resolved_ds,
                select_expression=select_expression,
                friendly_name=(
                    dedent(self.friendly_name) if self.friendly_name else self.friendly_name
                ),
                description=(dedent(self.description) if self.description else self.description),
                bigger_is_better=self.bigger_is_better,
                analysis_bases=self.analysis_bases
                or [AnalysisBasis.ENROLLMENTS, AnalysisBasis.EXPOSURES],
                type=self.type or "scalar",
                category=self.category,
                depends_on=upstream_metrics,
                owner=[self.owner] if isinstance(self.owner, str) else self.owner,
                deprecated=self.deprecated,
                level=self.level,
                analysis_units=analysis_units,
            )

        metrics_with_treatments = []

        if metric_summary:
            if self.statistics:
                for statistic_name, params in self.statistics.items():
                    stats_params = copy.deepcopy(params)
                    pre_treatments = []
                    for pt in stats_params.pop("pre_treatments", []):
                        if isinstance(pt, str):
                            ref = PreTreatmentReference(pt, {})
                        else:
                            name = pt.pop("name")
                            ref = PreTreatmentReference(name, pt)
                        pre_treatments.append(ref.resolve(spec))

                    metrics_with_treatments.append(
                        Summary(
                            metric=metric_summary[0].metric,
                            statistic=Statistic(statistic_name, stats_params),
                            pre_treatments=pre_treatments,
                        )
                    )
            else:
                metrics_with_treatments += metric_summary
        elif metric:
            if self.statistics is None:
                raise ValueError(f"No statistical treatment defined for metric '{self.name}'")

            for statistic_name, params in self.statistics.items():
                stats_params = copy.deepcopy(params)
                pre_treatments = []
                for pt in stats_params.pop("pre_treatments", []):
                    if isinstance(pt, str):
                        ref = PreTreatmentReference(pt, {})
                    else:
                        name = pt.pop("name")
                        ref = PreTreatmentReference(name, pt)
                    pre_treatments.append(ref.resolve(spec))

                metrics_with_treatments.append(
                    Summary(
                        metric=metric,
                        statistic=Statistic(statistic_name, stats_params),
                        pre_treatments=pre_treatments,
                    )
                )

        if len(metrics_with_treatments) == 0:
            raise ValueError(f"Metric {self.name} has no statistical treatment defined.")

        return metrics_with_treatments

    def merge(self, other: "MetricDefinition"):
        """Merge with another metric definition."""
        for key in attr.fields_dict(type(self)):
            setattr(self, key, getattr(other, key) or getattr(self, key))


MetricsConfigurationType = Dict[AnalysisPeriod, List[Summary]]


@attr.s(auto_attribs=True)
class MetricsSpec:
    """Describes the interface for the metrics section in configuration."""

    daily: List[MetricReference] = attr.Factory(list)
    weekly: List[MetricReference] = attr.Factory(list)
    days28: List[MetricReference] = attr.Factory(list)
    overall: List[MetricReference] = attr.Factory(list)
    preenrollment_weekly: List[MetricReference] = attr.Factory(list)
    preenrollment_days28: List[MetricReference] = attr.Factory(list)
    definitions: Dict[str, MetricDefinition] = attr.Factory(dict)

    @classmethod
    def from_dict(cls, d: dict) -> "MetricsSpec":
        params: Dict[str, Any] = {}
        known_keys = {f.name for f in attr.fields(cls)}
        for k in known_keys:
            if k == "days28":
                v = d.get("28_day", [])
            else:
                v = d.get(k, [])
            if not isinstance(v, list):
                raise ValueError(f"metrics.{k} should be a list of metrics")
            params[k] = [MetricReference(m) for m in v]

        params["definitions"] = {
            k: converter.structure(
                {"name": k, **dict((kk.lower(), vv) for kk, vv in v.items())},
                MetricDefinition,
            )
            for k, v in d.items()
            if k not in known_keys and k != "28_day"
        }

        return cls(**params)

    def resolve(
        self,
        spec: "AnalysisSpec",
        conf: Union["ExperimentConfiguration", "ProjectConfiguration"],
        configs: "ConfigCollection",
    ) -> MetricsConfigurationType:
        result = {}
        for period in AnalysisPeriod:
            summaries = [
                summary
                for ref in getattr(self, period.table_suffix)
                for summary in ref.resolve(spec, conf, configs)
            ]
            unique_summaries = []
            seen_summaries = set()

            # summaries needs to be reversed to make sure merged configs overwrite existing ones
            summaries.reverse()
            for summary in summaries:
                if (summary.metric.name, summary.statistic.name) not in seen_summaries:
                    seen_summaries.add((summary.metric.name, summary.statistic.name))
                    unique_summaries.append(summary)

            result[period] = unique_summaries

        return result

    def merge(self, other: "MetricsSpec"):
        """
        Merges another metrics spec into the current one.

        The `other` MetricsSpec overwrites existing metrics.
        """
        self.daily = other.daily + self.daily
        self.weekly = other.weekly + self.weekly
        self.days28 = other.days28 + self.days28
        self.overall = other.overall + self.overall
        self.preenrollment_weekly = other.preenrollment_weekly + self.preenrollment_weekly
        self.preenrollment_days28 = other.preenrollment_days28 + self.preenrollment_days28

        seen = set()
        for key, _ in self.definitions.items():
            for other_key in other.definitions:
                # support wildcard characters in `other`
                other_key_regex = re.compile(fnmatch.translate(other_key))
                if other_key_regex.fullmatch(key):
                    self.definitions[key].merge(other.definitions[other_key])
                    seen.add(other_key)
            seen.add(key)
        for key, definition in other.definitions.items():
            if key not in seen and is_valid_slug(key):
                self.definitions[key] = definition


converter.register_structure_hook(MetricsSpec, lambda obj, _type: MetricsSpec.from_dict(obj))



from textwrap import dedent

import pytest
import toml
from cattrs.errors import ClassValidationError

from metric_config_parser.metric import MetricReference
from metric_config_parser.monitoring import MonitoringSpec


class TestProjectSpec:
    def test_group_by_fail(self, config_collection):
        config_str = dedent(
            """
            [project]
            name = "foo"
            xaxis = "build_id"
            metrics = []

            [project.population]
            data_source = "foo"
            group_by_dimension = "os"

            [data_sources]
            [data_sources.foo]
            from_expression = "test"

            [dimensions]
            [dimensions.os]
            select_expression = "os"
            data_source = "foo"
            """
        )

        spec = MonitoringSpec.from_dict(toml.loads(config_str))

        with pytest.raises(ValueError):
            spec.resolve(experiment=None, configs=config_collection)

    def test_bad_project_dates(self):
        config_str = dedent(
            """
            [project]
            start_date = "My birthday"
            """
        )

        with pytest.raises(ClassValidationError):
            MonitoringSpec.from_dict(toml.loads(config_str))

    def test_bad_project_xaxis(self):
        config_str = dedent(
            """
            [project]
            xaxis = "Nothing"
            """
        )

        with pytest.raises(ClassValidationError):
            MonitoringSpec.from_dict(toml.loads(config_str))

    def test_metric_groups(self, config_collection):
        config_str = dedent(
            """
            [project]
            name = "foo"
            metrics = ["foo", "bar"]

            [project.metric_groups.foo]
            friendly_name = "Fooooo"
            metrics = [
                "foo",
                "bar"
            ]

            [project.population]
            data_source = "foo"

            [data_sources]
            [data_sources.foo]
            from_expression = "test"

            [metrics]
            [metrics.foo]
            select_expression = "foo"
            data_source = "foo"

            [metrics.foo.statistics]
            sum = {}

            [metrics.bar]
            select_expression = "bar"
            data_source = "foo"

            [metrics.bar.statistics]
            sum = {}
            """
        )

        spec = MonitoringSpec.from_dict(toml.loads(config_str))
        config = spec.resolve(experiment=None, configs=config_collection)

        assert len(config.project.metric_groups) == 1
        assert config.project.metric_groups[0].name == "foo"
        assert MetricReference("foo") in config.project.metric_groups[0].metrics
        assert MetricReference("bar") in config.project.metric_groups[0].metrics
